/** 
 *   Camera Control for STAR I.
 *
 *   (c) Copyright 2005 Federico Bareilles <fede@iar.unlp.edu.ar>,
 *   Instituto Argentino de Radio Astronomia (IAR).
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *     
 *   The author does NOT admit liability nor provide warranty for any
 *   of this software. This material is provided "AS-IS" in the hope
 *   that it may be useful for others.
 *
 **/
#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
//#include <stdarg.h>
#include <limits.h>

#include "fitsio.h"
#include "my_stdio.h"
#include "local_gpib.h"
#include "camera.h"
#include "camera_setup.h"
#include "client/camera_tools.h"

/* Local emulated function: */
static int e_ibclr   ( int ud );
static int e_ibconfig( int ud, int option, int value );
static int e_ibdev   ( int board_index, int pad, int sad, 
		int timo, int send_eoi, int eosmode );
static int e_ibln    ( int ud, int pad, int sad, short *found_listener );
static int e_ibrd    ( int ud, void *buf, long count );
static int e_ibrsp   ( int ud, char *spr );
static int e_ibtmo   ( int ud, int v );
static int e_ibwrt   ( int ud, const void *buf, long count );
static int e_ibwrta  ( int ud, const void *buf, long count );


/* Exported functions: */
int (*l_ibclr)   ( int ud ) = e_ibclr ;
int (*l_ibconfig)( int ud, int option, int value ) = e_ibconfig;
int (*l_ibdev)   ( int board_index, int pad, int sad, 
		   int timo, int send_eoi, int eosmode ) = e_ibdev;
int (*l_ibln)    ( int ud, int pad, int sad, short *found_listener ) = e_ibln;
int (*l_ibrd)    ( int ud, void *buf, long count ) = e_ibrd;
int (*l_ibrsp)   ( int ud, char *spr ) = e_ibrsp;
int (*l_ibtmo)   ( int ud, int v ) = e_ibtmo;
int (*l_ibwrt)   ( int ud, const void *buf, long count ) = e_ibwrt;
int (*l_ibwrta)  ( int ud, const void *buf, long count ) = e_ibwrta;


static int read_config_file(char *filename);
volatile int ibsta = 0,
	ibcnt = 0,
	iberr = 0;

#define BUFFER_LEN 16
static char result[BUFFER_LEN];
static int read_lut_flag = 0;
static int send_data_flag = 0;

static int etime = 2;
static int letime = 2;
static int treg = 1;
static int temperature = -45;
static int gain = 0;
static int xx0 = 0;
static int xxlen = 576;
static int yy0 = 0;
static int yylen = 384;
static int auto_win = 1;
static int max_win = 254;
static int min_win = 10;
static int zoom = 0;
static char *comment = NULL;
static int mouse_x = 0;
static int mouse_y = 0;
static int mouse_v = 2048;
static int imtype = IMT_NONE;

static char elocation_path[_POSIX_PATH_MAX];
static char eexpose_file[_POSIX_PATH_MAX];
static char edark_file[_POSIX_PATH_MAX];
static char ebias_file[_POSIX_PATH_MAX];


/* FIXME: */
static int e_ibclr( int ud )
{
	PVERB(PINFO, "ibclr:\n");
	return 0;
}


static int e_ibconfig( int ud, int option, int value )
{
	PVERB(PINFO, "ibconfig:\n");

	ibsta = 0;

	return ibsta;
}


/* FIXME: */
static int e_ibdev( int board_index, int pad, int sad, 
		int timo, int send_eoi, int eosmode )
{
	PVERB(PINFO, "ibdev:\n");

	read_config_file(EMULATOR_CONFIG_FILE);

	return 1;
}


/* FIXME: */
static int e_ibln( int ud, int pad, int sad, short *found_listener )
{
	PVERB(PINFO, "ibln:\n");

	return 0;
}


static int read_scale_fits(int imtype, char *image, void * buf, 
			   int y0, int ylen,
			   int x0, int xlen )
{
	int i, j=0;
	long count = (long) (((ylen - y0)*(xlen-x0)) << 1);
	/*
	for(i=0; i < (count>>1);i++){
		if (j == 384) j = 0;
		*(((int16_t *)buf)+i) = j & 0xfff;
		j++;
	}
	*/
	/* FITS file pointer, defined in fitsio.h */
	fitsfile *fptr;
	/* CFITSIO status value MUST be initialized to zero! */   
	int status = 0;
	int bitpix, naxis;
	long naxes[2] = {1,1}, fpixel[2] = {1,1};
	double *pixels;
	double factor = 1.0;
	double exposure;
	char file[_POSIX_PATH_MAX];
	int16_t *bias = NULL;
	int file_gain;
	unsigned int value = 0;

	sprintf(file, "%s[%d:%d,%d:%d]", image, y0 + 1, ylen, x0 + 1, xlen);


	if (!fits_open_file(&fptr, file, READONLY, &status)) {
		if ( fits_read_key(fptr, TDOUBLE, "EXPTIME", &exposure, NULL,
				   &status) ) {
			PVERB(PERR, "getting %s: EXPTIME not found\n", file);
			return -1;
		}
		if ( fits_read_key(fptr, TINT, "GAIN", &file_gain, NULL,
				   &status) ) {
			PVERB(PERR, "getting %s: GAIN not found\n", file);
			return -1;
		}
		if (imtype == IMT_IMAGE || imtype == IMT_DARK) {
			factor =  ((double)letime * (double)(gain * 3 + 1)) / 
				((double)exposure * (double)file_gain * 10.0);
			//printf("*** factor: %f\n", factor);
			sprintf(file, "%s/%s", elocation_path, ebias_file);
			bias = (int16_t *) malloc((count>>1)*sizeof(int16_t));
			read_scale_fits(IMT_BIAS, 
					file, bias, y0, ylen, x0, xlen);
		}
		if (!fits_get_img_param(fptr, 2, &bitpix, &naxis, naxes, 
					&status) ) {
			if (naxis > 2 || naxis == 0) {
				PVERB(PERR,
				"Error: only 1D or 2D images are supported\n");
			} else {
				/* get memory for 1 row */
				pixels = (double *) malloc(naxes[0] * 
							   sizeof(double));
				if (pixels == NULL) {
					PVERB(PERR,
					      "Memory allocation error\n");
					return -1;
				}
				j = count >> 1;
				/* loop over all the rows in the
                                   image, top to bottom */
				for (fpixel[1] = naxes[1]; fpixel[1] >= 1;
				     fpixel[1]--) {
					/* read row of pixels */
					if (fits_read_pix(fptr, TDOUBLE, 
							  fpixel, naxes[0], 
							  NULL, pixels, NULL, 
							  &status) )
						break;  /* jump out of loop on error */
					for (i = naxes[0] -1 ; i >= 0; i--) {
						j--;
						if ( imtype == IMT_IMAGE || 
						     imtype == IMT_DARK) {
							value = ((unsigned int) ((pixels[i]-bias[j]) * factor) + bias[j]);
							if (value > 0x0fff) {
								value = 0x0fff;
							}
							*(((int16_t *)buf)+j) = (int16_t) value;
						} else {
							value =  ((unsigned int) pixels[i]) & 0x0fff;
						}
						*(((int16_t *)buf)+j) = (int16_t) value;
					}
				}
				free(pixels);
				if ( imtype == IMT_IMAGE || imtype == IMT_DARK)
					free(bias);
			
			}
		}
		fits_close_file(fptr, &status);
	}
	/* print any error message */
	if (status) fits_report_error(stderr, status);
	return(status);
	
	return 0;
}


/* FIXME: */
static int e_ibrd( int ud, void *buf, long count )
{
	int i, j=0;
	char image[_POSIX_PATH_MAX];

	PVERB(PINFO, "ibrd:\n");
	if ( send_data_flag ) {
		send_data_flag = 0;
		if ( ((yylen-yy0) * (xxlen-xx0)) != (count >> 1)) {
			PVERB(PERR, "Dimension not match\n");
			return -1;
		}
		switch(imtype) {
		case IMT_IMAGE:
			sprintf(image, "%s/%s", elocation_path, eexpose_file);
			read_scale_fits(IMT_IMAGE, 
					image, buf, yy0, yylen, xx0, xxlen);
			break;
		case IMT_DARK:
			sprintf(image, "%s/%s", elocation_path, edark_file);
			read_scale_fits(IMT_DARK,
					image, buf, yy0, yylen, xx0, xxlen);
			break;
		case IMT_BIAS:
			sprintf(image, "%s/%s", elocation_path, ebias_file);
			read_scale_fits(IMT_BIAS, 
					image, buf, yy0, yylen, xx0, xxlen);
			break;
		case IMT_CLEAR:
			for(i=0; i < (count>>1);i++){
				*(((char *) buf)+(i<<1)) = 0;
				*(((char *) buf)+(i<<1|1)) = 0;
			}
			break;
		case IMT_NONE:
		default:
			for(i=0; i < (count>>1);i++){
				if (j == 384) j = 0;
				*(((char *) buf)+(i<<1)) = j & 0xff;
				*(((char *) buf)+(i<<1|1)) = (j & 0xf00) >> 8;
				j++;
			}
			break;
		}
	}else {
		memcpy(buf, result, count);
	}
	ibcnt = count;

	return 0;
}


/* FIXME: */
static int e_ibrsp( int ud, char *spr )
{
	PVERB(PINFO, "ibrsp:");
	*spr = CMD_OK;

	return 0;
}


/* FIXME: */
static int e_ibtmo( int ud, int v )
{
	PVERB(PINFO, "ibtmo:");

	return 0;
}


/* FIXME: */
static int e_ibwrt( int ud, const void *buf, long count )
{
	char type = *((char *)buf);

	PVERB(PINFO, "ibwrt: %s [%ld]\n", (char *)buf, count);

	if ( read_lut_flag ) {
		read_lut_flag = 0;
		//printf("Rcv LUT: %ld bytes\n", count);
		return 0;
	}
	switch(type) {
	case '@': /* Read command */
		memset(&result, 0, BUFFER_LEN);
		switch( *((char *)buf+1) ) {
		case 'A': sprintf(result,"%d",etime); break;
		case 'B': sprintf(result,"%d",gain);  break;
		case 'C': sprintf(result,"%d",treg);  break;
		case 'D': sprintf(result,"%d",yy0);    break;
		case 'E': sprintf(result,"%d",xx0);    break;
		case 'F': sprintf(result,"%d",yylen);  break;
		case 'G': sprintf(result,"%d",xxlen);  break;
		case 'H': sprintf(result,"%d",zoom);  break;
		case 'I': sprintf(result,"%d",auto_win); break;
		case 'J': if ( comment ) sprintf(result,"%s",comment); break;
		case 'K':
		case 'L':
		case 'M': PVERB(PERR,"Not implement (%s)\n",(char *)buf);
			break;
		case 'N': sprintf(result,"%d",max_win);  break;
		case 'O': sprintf(result,"%d",min_win);  break;
		case 'P':
		case 'Q':
		case 'R':
		case 'S': PVERB(PERR,"Not implement (%s)\n",(char *)buf);
			break;
		case 'T': sprintf(result,"%d",temperature); break;
		case 'U': sprintf(result,"%d",mouse_y);  break;
		case 'V': sprintf(result,"%d",mouse_x);  break;
		case 'W': sprintf(result,"%d",mouse_v);  break;
		default:
			PVERB(PERR, "Unknow R cmd: %s\n", (char *) buf);
			break;
		}
		break;
	case '!': /* Write command */
		switch( *((char *)buf+1) ) {
		case 'A': etime = atoi( ((char *)buf+2) );    break;
		case 'B': gain = atoi( ((char *)buf+2) );     break;
		case 'C': treg = atoi( ((char *)buf+2) );     break;
		case 'D': yy0 = atoi( ((char *)buf+2) );       break;
		case 'E': xx0 = atoi( ((char *)buf+2) );       break;
		case 'F': yylen = atoi( ((char *)buf+2) );     break;
		case 'G': xxlen = atoi( ((char *)buf+2) );     break;
		case 'H': zoom = atoi( ((char *)buf+2) );     break;
		case 'I': auto_win = atoi( ((char *)buf+2) ); break;
		case 'J':
			if ( comment ) free(comment);
			comment = (char *) malloc((count - 2) * sizeof(char));
			memcpy(comment, ((char *)buf+2), count - 2);
			break;
		case 'K':
		case 'L':
		case 'M': PVERB(PERR,"Not implement (%s)\n",(char *)buf);
			break;
		case 'N': max_win = atoi( ((char *)buf+2) );  break;
		case 'O': min_win = atoi( ((char *)buf+2) );  break;
		case 'P':
		case 'Q':
		case 'R':
		case 'S': PVERB(PERR,"Not implement (%s)\n",(char *)buf);
			break;
		case 'U': mouse_y = atoi( ((char *)buf+2) );  break;
		case 'V': mouse_x = atoi( ((char *)buf+2) );  break;
		case 'W': mouse_v = atoi( ((char *)buf+2) );  break;
		default:
			PVERB(PERR, "Unknow W cmd: %s\n", (char *) buf);
			break;
		}
		break;
	case ':': /* Action command */
		switch( *((char *)buf+1) ) {
		case 'A': imtype = IMT_IMAGE; letime = etime; break;
		case 'B': imtype = IMT_DARK;  letime = etime; break;
		case 'C': imtype = IMT_BIAS;  break;
		case 'G': imtype = IMT_SHADE; break;
		case 'H': imtype = IMT_CLEAR; break;

		case 'J':
			send_data_flag = 1;
			break;

		case 'R': /* FIXME */break;
		case 'U': 
			read_lut_flag = 1;
			break;
		default:
			PVERB(PERR, "Unknow Action cmd: %s\n", (char *) buf);
			break;
		}
		break;
	default:
		//iberr = 1;
		PVERB(PERR, "Unknow cmd: %s\n", (char *) buf);
	}

	return 0;
}


static int e_ibwrta( int ud, const void *buf, long count )
{
	PVERB(PINFO, "ibwrta:");

	return e_ibwrt( ud, buf, count );
}



static int read_config_file(char *filename)
{
	FILE *fp;
	char line[LINE_LEN_CONF];
	char *sline, *tag, *data, *comment;

	fp = fopen(filename, "r");
	
	if( fp == NULL ){
                SYSLOG(LOG_ERR,
                       "read_config_file: Error opening file: %s",
		       filename);
                return -1;
        }
	
	memset(line, 0, sizeof(line));
	while( fgets(line, sizeof(line), fp)){
		if(line[strlen(line)-1] == '\n')
			line[strlen(line)-1] = 0;
		sline = stripwhite(line);
		if (parse_line(sline, &tag, &data, &comment) == 0){
			if( memcmp(tag,"ELOCATION", 9) == 0 ){
				sprintf(elocation_path, "%s", data);
			}else if( memcmp(tag,"EEXPOSE", 7) == 0 ){
				sprintf(eexpose_file, "%s", data);
			}else if( memcmp(tag,"EDARK", 5) == 0 ){
				sprintf(edark_file, "%s", data);
			}else if( memcmp(tag,"EBIAS", 5) == 0 ){
				sprintf(ebias_file, "%s", data);
			}else{
				SYSLOG(LOG_ERR, "Bad parameter(%s) in %s.",
				       line, filename);
			}
		}
		memset(line, 0, sizeof(line));
	}
	fclose(fp);

	return 0;
}

